package movies

import (
	"fmt"

	"gitlab.quimbo.fr/odwrtw/canape-sql/src/internal/users"
	"gitlab.quimbo.fr/odwrtw/canape-sql/src/internal/web"
)

const (
	upsertWishlistQuery = `
	INSERT INTO movies_tracked (imdb_id, user_id)
	VALUES ($1, $2)
	ON CONFLICT (imdb_id, user_id)
	DO UPDATE
	SET imdb_id=$1, user_id=$2;`

	getWishlistQueryByUserID = `
	SELECT
	movies.id,
	movies.title,
	movies.imdb_id,
	movies.tmdb_id,
	movies.votes,
	movies.rating,
	movies.plot,
	movies.year,
	movies.original_title,
	movies.runtime,
	movies.genres,
	movies.sort_title,
	movies.tagline,
	movies.created_at,
	movies.updated_at,
	movies_tracked.user_id
	FROM movies INNER JOIN movies_tracked
	ON movies.imdb_id=movies_tracked.imdb_id AND movies_tracked.user_id=$1;`

	deleteWishlistedQueryByID = `DELETE FROM movies_tracked WHERE imdb_id=$1 AND user_id=$2;`
)

// AddToWishlist Adds a movie to a user's wishlist
func (m *Movie) AddToWishlist(env *web.Env, user *users.User) error {
	_, err := env.Database.Exec(upsertWishlistQuery, m.ImdbID, user.ID)
	if err != nil {
		return err
	}
	return nil
}

// DeleteFromWishlist deletes a movie from a user's wishlist
func (m *Movie) DeleteFromWishlist(env *web.Env, user *users.User) error {
	r, err := env.Database.Exec(deleteWishlistedQueryByID, m.ImdbID, user.ID)
	if err != nil {
		return err
	}
	count, _ := r.RowsAffected()
	if count != 1 {
		return fmt.Errorf("Unexpected number of row deleted: %d", count)
	}
	return nil
}

// GetWishlist returns a list of movies wishlisted by user
func GetWishlist(env *web.Env, user *users.User) ([]*Movie, error) {
	var moviesDB = []*MovieDB{}
	err := env.Database.Select(&moviesDB, getWishlistQueryByUserID, user.ID)
	if err != nil {
		return nil, err
	}

	var movies []*Movie
	for _, movieDB := range moviesDB {
		movie := New(movieDB.ImdbID)
		// Set the poster url
		movie.PosterURL = movie.GetPosterURL(env)

		movie.FillFromDB(movieDB)
		movies = append(movies, movie)
	}

	return movies, nil
}
