package shows

import (
	"database/sql"
	"time"

	"github.com/Sirupsen/logrus"
	"github.com/jmoiron/sqlx"
	polochon "github.com/odwrtw/polochon/lib"
	"gitlab.quimbo.fr/odwrtw/canape-sql/src/internal/sqly"
	"gitlab.quimbo.fr/odwrtw/canape-sql/src/internal/torrents"
	"gitlab.quimbo.fr/odwrtw/canape-sql/src/internal/web"
)

const (
	upsertEpisodeQuery = `
	INSERT INTO episodes (show_imdb_id, show_tvdb_id, title, season, episode, tvdb_id, aired, plot, runtime, rating, imdb_id)
	VALUES (:show_imdb_id, :show_tvdb_id, :title, :season, :episode, :tvdb_id, :aired, :plot, :runtime, :rating, :imdb_id)
	ON CONFLICT (show_imdb_id, season, episode)
	DO UPDATE
	SET show_imdb_id=:show_imdb_id, show_tvdb_id=:show_tvdb_id, title=:title,
	season=:season, episode=:episode, tvdb_id=:tvdb_id, aired=:aired,
	plot=:plot, runtime=:runtime, rating=:rating, imdb_id=:imdb_id
	RETURNING id;`

	getEpisodesQuery = `
	SELECT *
	FROM episodes WHERE show_imdb_id=$1;`
)

// Episode represents an episode
type Episode struct {
	sqly.BaseModel
	polochon.ShowEpisode
	PolochonURL string `json:"polochon_url"`
}

// EpisodeDB represents the Episode in the DB
type EpisodeDB struct {
	ID           string    `db:"id"`
	TvdbID       int       `db:"tvdb_id"`
	ImdbID       string    `db:"imdb_id"`
	ShowImdbID   string    `db:"show_imdb_id"`
	ShowTvdbID   int       `db:"show_tvdb_id"`
	Season       int       `db:"season"`
	Episode      int       `db:"episode"`
	Title        string    `db:"title"`
	Rating       float32   `db:"rating"`
	Plot         string    `db:"plot"`
	Thumb        string    `db:"thumb"`
	Runtime      int       `db:"runtime"`
	Aired        string    `db:"aired"`
	ReleaseGroup string    `db:"release_group"`
	Created      time.Time `db:"created_at"`
	Updated      time.Time `db:"updated_at"`
}

// NewEpisode returns an Episode
func NewEpisode() *Episode {
	return &Episode{}
}

// NewEpisodeDB returns an Episode ready to be put in DB from an
// Episode
func NewEpisodeDB(e *Episode) EpisodeDB {
	return EpisodeDB{
		ID:           e.ID,
		TvdbID:       e.TvdbID,
		ImdbID:       e.EpisodeImdbID,
		ShowImdbID:   e.ShowImdbID,
		ShowTvdbID:   e.ShowTvdbID,
		Season:       e.Season,
		Episode:      e.Episode,
		Title:        e.Title,
		Rating:       e.Rating,
		Plot:         e.Plot,
		Thumb:        e.Thumb,
		Runtime:      e.Runtime,
		Aired:        e.Aired,
		ReleaseGroup: e.ReleaseGroup,
		Created:      e.Created,
		Updated:      e.Updated,
	}
}

// FillFromDB returns a Show from a ShowDB extracted from the DB
func (e *Episode) FillFromDB(eDB *EpisodeDB) {
	e.ID = eDB.ID
	e.TvdbID = eDB.TvdbID
	e.EpisodeImdbID = eDB.ImdbID
	e.ShowImdbID = eDB.ShowImdbID
	e.ShowTvdbID = eDB.ShowTvdbID
	e.Season = eDB.Season
	e.Episode = eDB.Episode
	e.Title = eDB.Title
	e.Rating = eDB.Rating
	e.Plot = eDB.Plot
	e.Thumb = eDB.Thumb
	e.Runtime = eDB.Runtime
	e.Aired = eDB.Aired
	e.Created = eDB.Created
	e.Updated = eDB.Updated
}

// Upsert episode to the database
func (e *Episode) Upsert(db *sqlx.DB) error {
	eDB := NewEpisodeDB(e)
	var id string
	r, err := db.NamedQuery(upsertEpisodeQuery, eDB)
	if err != nil {
		return err
	}
	for r.Next() {
		r.Scan(&id)
	}
	e.ID = id
	return nil
}

// GetTorrents retrieves torrents for the show, first try to get info from db,
// if not exists, use polochon.Torrenter and save informations in the database
// for future use
//
// If force is used, the torrenter will be used even if the torrent is found in
// database
func (e *Episode) GetTorrents(env *web.Env, force bool) error {
	log := env.Log.WithFields(logrus.Fields{
		"imdb_id":  e.ShowEpisode.ShowImdbID,
		"season":   e.ShowEpisode.Season,
		"episode":  e.ShowEpisode.Episode,
		"function": "shows.GetTorrents",
	})
	log.Debugf("getting torrents")

	if len(e.Torrenters) == 0 {
		e.Torrenters = env.Config.ShowTorrenters
	}

	episodeTorrents, err := torrents.GetEpisodeTorrents(
		env.Database,
		e.ShowEpisode.ShowImdbID,
		e.ShowEpisode.Season,
		e.ShowEpisode.Episode,
	)
	switch err {
	case nil:
		log.Debug("torrents found in database")
	case sql.ErrNoRows:
		log.Debug("torrent not found in database")
		// We'll need to GetTorrents from torrenters
	default:
		// Unexpected error
		return err
	}
	if !force {
		log.Debugf("returning %d torrents from db", len(episodeTorrents))
		// Add the torrents to the episode
		for _, t := range episodeTorrents {
			e.Torrents = append(e.Torrents, t.Torrent)
		}
		return nil
	}

	err = e.ShowEpisode.GetTorrents(env.Log)
	if err != nil {
		return err
	}

	log.Debugf("got %d torrents from torrenters", len(e.ShowEpisode.Torrents))

	for _, t := range e.ShowEpisode.Torrents {
		torrent := torrents.NewEpisodeFromPolochon(e.ShowEpisode, t)
		err = torrent.Upsert(env.Database)
		if err != nil {
			log.Errorf("error while adding torrent : %s", err)
			continue
		}
	}

	return nil
}
