package shows

import (
	"fmt"

	"gitlab.quimbo.fr/odwrtw/canape-sql/src/internal/users"
	"gitlab.quimbo.fr/odwrtw/canape-sql/src/internal/web"
)

const (
	upsertWishlistQuery = `
	INSERT INTO shows_tracked (imdb_id, user_id, season, episode)
	VALUES ($1, $2, $3, $4)
	ON CONFLICT (imdb_id, user_id)
	DO UPDATE
	SET imdb_id=$1, user_id=$2, season=$3, episode=$4;`

	getWishlistQueryByUserID = `
	SELECT
	shows.id,
	shows.imdb_id,
	shows.title,
	shows.rating,
	shows.plot,
	shows.tvdb_id,
	shows.year,
	shows.first_aired,
	shows_tracked.season,
	shows_tracked.episode,
	shows.updated_at,
	shows.created_at
	FROM shows INNER JOIN shows_tracked
	ON shows.imdb_id=shows_tracked.imdb_id AND shows_tracked.user_id=$1;`

	deleteWishlistedQueryByID = `DELETE FROM shows_tracked WHERE imdb_id=$1 AND user_id=$2;`
)

// AddToWishlist Adds a show to a user's wishlist
func (s *Show) AddToWishlist(env *web.Env, user *users.User, season, episode int) error {
	_, err := env.Database.Exec(upsertWishlistQuery, s.ImdbID, user.ID, season, episode)
	if err != nil {
		return err
	}
	return nil
}

// IsTracked returns true if the show is tracked use this function
// after retrieve the show with GetAsUser or other *AsUser functions
func (s *Show) IsTracked() bool {
	if s.TrackedSeason != nil && s.TrackedEpisode != nil {
		return true
	}
	return false
}

// DeleteFromWishlist deletes a show from a user's wishlist
func (s *Show) DeleteFromWishlist(env *web.Env, user *users.User) error {
	r, err := env.Database.Exec(deleteWishlistedQueryByID, s.ImdbID, user.ID)
	if err != nil {
		return err
	}
	count, _ := r.RowsAffected()
	if count != 1 {
		return fmt.Errorf("Unexpected number of row deleted: %d", count)
	}
	return nil
}

// GetWishlist returns a list of shows tracked by user
func GetWishlist(env *web.Env, user *users.User) ([]*Show, error) {
	var showsDB = []*ShowDB{}
	err := env.Database.Select(&showsDB, getWishlistQueryByUserID, user.ID)
	if err != nil {
		return nil, err
	}

	var shows []*Show
	for _, showDB := range showsDB {
		// Set the poster url
		show := New(showDB.ImdbID)
		show.PosterURL = show.GetPosterURL(env)
		show.FillFromDB(showDB)
		shows = append(shows, show)
	}

	return shows, nil
}
