package torrents

import (
	"database/sql"
	"fmt"
	"time"

	"gitlab.quimbo.fr/odwrtw/canape-sql/src/internal/sqly"

	"github.com/jmoiron/sqlx"
	polochon "github.com/odwrtw/polochon/lib"
)

const (
	upsertMovieTorrentQuery = `
	INSERT INTO movie_torrents (imdb_id, url, source, quality, upload_user,
	seeders, leechers)
	VALUES (:imdb_id, :url, :source, :quality, :upload_user, :seeders,
	:leechers)
	ON CONFLICT (imdb_id, quality, source)
	DO UPDATE SET imdb_id=:imdb_id, url=:url, source=:source, quality=:quality,
	upload_user=:upload_user, seeders=:seeders, leechers=:leechers
	RETURNING id;`

	getMovieTorrentQueryByImdbID = `
	SELECT *
	FROM movie_torrents WHERE imdb_id=$1;`

	getMovieTorrentQueryByID = `
	SELECT *
	FROM movie_torrents WHERE id=$1;`

	deleteMovieTorrentQuery = `DELETE FROM movie_torrents WHERE id=$1;`
)

// MovieTorrent represents a movie torrent
type MovieTorrent struct {
	sqly.BaseModel
	polochon.Torrent
	ImdbID string `json:"imdb_id"`
}

// MovieTorrentDB represents the MovieTorrent in the DB
type MovieTorrentDB struct {
	ID         string    `db:"id"`
	ImdbID     string    `db:"imdb_id"`
	URL        string    `db:"url"`
	Source     string    `db:"source"`
	Quality    string    `db:"quality"`
	UploadUser string    `db:"upload_user"`
	Seeders    int       `db:"seeders"`
	Leechers   int       `db:"leechers"`
	Created    time.Time `db:"created_at"`
	Updated    time.Time `db:"updated_at"`
}

// NewMovieFromImDB returns a new MovieTorrent with an ImDB id
func NewMovieFromImDB(imdbID string) *MovieTorrent {
	return &MovieTorrent{
		ImdbID: imdbID,
	}
}

// NewMovie returns a new MovieTorrent with an ImDB id
func NewMovie(imdbID string, poTo polochon.Torrent) *MovieTorrent {
	return &MovieTorrent{
		ImdbID:  imdbID,
		Torrent: poTo,
	}
}

// Get returns show details in database from id or imdbid or an error
func (m *MovieTorrent) Get(db *sqlx.DB) error {
	var mDB MovieTorrentDB
	var err error
	if m.ID != "" {
		err = db.QueryRowx(getMovieTorrentQueryByID, m.ID).StructScan(&mDB)
	} else if m.ImdbID != "" {
		err = db.QueryRowx(getMovieTorrentQueryByImdbID, m.ImdbID).StructScan(&mDB)
	} else {
		err = fmt.Errorf("can't get movie torrent details, you have to specify an ID or ImdbID")
	}
	if err != nil {
		return err
	}
	m.FillFromDB(&mDB)
	return nil
}

// GetMovieTorrents returns show details in database from id or imdbid or an error
func GetMovieTorrents(db *sqlx.DB, imdbID string) ([]*MovieTorrent, error) {
	var torrentsDB = []*MovieTorrentDB{}
	err := db.Select(&torrentsDB, getMovieTorrentQueryByImdbID, imdbID)
	if err != nil {
		return nil, err
	}

	if len(torrentsDB) == 0 {
		return nil, sql.ErrNoRows
	}

	var torrents []*MovieTorrent
	for _, torrentDB := range torrentsDB {
		movie := NewMovieFromImDB(imdbID)
		movie.FillFromDB(torrentDB)
		torrents = append(torrents, movie)
	}

	return torrents, nil
}

// Upsert a movie torrent in the database
func (m *MovieTorrent) Upsert(db *sqlx.DB) error {
	mDB := NewMovieTorrentDB(m)
	var id string
	r, err := db.NamedQuery(upsertMovieTorrentQuery, mDB)
	if err != nil {
		return err
	}
	for r.Next() {
		r.Scan(&id)
	}
	m.ID = id

	return nil
}

// Delete movie from database
func (m *MovieTorrent) Delete(db *sqlx.DB) error {
	r, err := db.Exec(deleteMovieTorrentQuery, m.ID)
	if err != nil {
		return err
	}
	count, _ := r.RowsAffected()
	if count != 1 {
		return fmt.Errorf("Unexpected number of row deleted: %d", count)
	}
	return nil
}

// NewMovieTorrentDB returns a MovieTorrent ready to be put in DB from a
// MovieTorrent
func NewMovieTorrentDB(m *MovieTorrent) MovieTorrentDB {
	return MovieTorrentDB{
		ID:         m.ID,
		ImdbID:     m.ImdbID,
		URL:        m.URL,
		Source:     m.Source,
		Quality:    string(m.Quality),
		UploadUser: m.UploadUser,
		Seeders:    m.Seeders,
		Leechers:   m.Leechers,
		Created:    m.Created,
		Updated:    m.Updated,
	}
}

// FillFromDB will fill a MovieTorrent from a MovieTorrentDB extracted from the DB
func (m *MovieTorrent) FillFromDB(mDB *MovieTorrentDB) {
	q, _ := polochon.StringToQuality(mDB.Quality)
	m.ImdbID = mDB.ImdbID
	m.Created = mDB.Created
	m.Updated = mDB.Updated
	m.ID = mDB.ID
	m.URL = mDB.URL
	m.Source = mDB.Source
	m.Quality = *q
	m.UploadUser = mDB.UploadUser
	m.Seeders = mDB.Seeders
	m.Leechers = mDB.Leechers
}
